//
// Created by philipp on 01.12.17.
//

#ifndef PROTOCOL_PLATOONCREATEMESSAGE_H
#define PROTOCOL_PLATOONCREATEMESSAGE_H

#include <cstdint>
#include <vector>
#include "Protocol.h"
#include <NetworkingLib/Message.h>

namespace protocol
{

namespace messageTypes
{
constexpr MessageType LEADER_REQUEST = 0x01;
constexpr MessageType FOLLOWER_REQUEST = 0x02;
constexpr MessageType ACCEPT_RESPONSE = 0x03;
constexpr MessageType REJECT_RESPONSE = 0x04;
constexpr MessageType HEARTBEAT = 0x05;
constexpr MessageType BROADCAST = 0x06;
constexpr MessageType LEAVE_PLATOON = 0x07;
}

union FloatConverter
{
    float floatVal;
    char bytes[4];
};

class PlatoonMessage
{
public:
    static constexpr auto SIZE = sizeof(VehicleId)
                                 + sizeof(MessageType)
                                 + sizeof(PlatoonId)
                                 + sizeof(PlatoonSpeed)
                                 + sizeof(InnerPlatoonDistance);

    PlatoonMessage()
        : valid{false}
    {}

    PlatoonMessage(
        VehicleId vehicleId,
        MessageType messageType,
        PlatoonId platoonId,
        PlatoonSpeed platoonSpeed,
        InnerPlatoonDistance innerPlatoonDistance)
        : vehicleId(vehicleId)
          , messageType(messageType)
          , platoonId(platoonId)
          , platoonSpeed(platoonSpeed)
          , innerPlatoonDistance(innerPlatoonDistance)
          , valid{true}
    {}

    static PlatoonMessage followerRequest(VehicleId vehicleId)
    { return PlatoonMessage(vehicleId, messageTypes::FOLLOWER_REQUEST, 0, 0, 0); }

    static PlatoonMessage leaderRequest(VehicleId vehicleId, PlatoonId platoonId)
    { return PlatoonMessage(vehicleId, messageTypes::LEADER_REQUEST, platoonId, 0, 0); }

    static PlatoonMessage rejectResponse(VehicleId vehicleId)
    { return PlatoonMessage(vehicleId, messageTypes::REJECT_RESPONSE, 0, 0, 0); }

    static PlatoonMessage acceptResponse(VehicleId vehicleId, PlatoonId platoonId)
    { return PlatoonMessage(vehicleId, messageTypes::ACCEPT_RESPONSE, platoonId, 0, 0); }

    static PlatoonMessage broadcastMessage(VehicleId vehicleId,
                                           PlatoonId platoonId,
                                           PlatoonSpeed platoonSpeed,
                                           InnerPlatoonDistance innerPlatoonDistance)
    { return PlatoonMessage(vehicleId, messageTypes::BROADCAST, platoonId, platoonSpeed, innerPlatoonDistance); }

    static PlatoonMessage heartbeatMessage(VehicleId vehicleId, PlatoonId platoonId)
    { return PlatoonMessage(vehicleId, messageTypes::HEARTBEAT, platoonId, 0, 0); }

    static PlatoonMessage leaveMessage(VehicleId vehicleId, PlatoonId platoonId)
    { return PlatoonMessage(vehicleId, messageTypes::LEAVE_PLATOON, platoonId, 0, 0); }

    MessageType getMessageType() const noexcept
    { return messageType; }

    VehicleId getVehicleId() const noexcept
    { return vehicleId; }

    PlatoonId getPlatoonId() const noexcept
    { return platoonId; }

    PlatoonSpeed getPlatoonSpeed() const noexcept
    { return platoonSpeed; }

    InnerPlatoonDistance getInnerPlatoonDistance() const noexcept
    { return innerPlatoonDistance; }

    bool isValid() const noexcept
    { return valid; }

private:
    VehicleId vehicleId;
    MessageType messageType;
    PlatoonId platoonId;
    PlatoonSpeed platoonSpeed;
    InnerPlatoonDistance innerPlatoonDistance;

    bool valid;
};
}

namespace networking
{
namespace message
{

template<>
struct Encoder<protocol::PlatoonMessage>
{
    std::string operator()(const protocol::PlatoonMessage & message)
    {
        using namespace protocol;

        auto vehicleId = message.getVehicleId();
        auto messageType = message.getMessageType();
        auto platoonId = message.getPlatoonId();
        auto platoonSpeed = message.getPlatoonSpeed();
        auto innerPlatoonDistance = message.getInnerPlatoonDistance();

        std::string data(PlatoonMessage::SIZE, '\0');

        data[0] = (std::uint8_t) (vehicleId & 0x000000ff);
        data[1] = (std::uint8_t) ((vehicleId & 0x0000ff00) >> 8);
        data[2] = (std::uint8_t) ((vehicleId & 0x00ff0000) >> 16);
        data[3] = (std::uint8_t) ((vehicleId & 0xff000000) >> 24);
        data[4] = (std::uint8_t) (messageType);
        data[5] = (std::uint8_t) (platoonId & 0x000000ff);
        data[6] = (std::uint8_t) ((platoonId & 0x0000ff00) >> 8);
        data[7] = (std::uint8_t) ((platoonId & 0x00ff0000) >> 16);
        data[8] = (std::uint8_t) ((platoonId & 0xff000000) >> 24);
        FloatConverter converter;
        converter.floatVal = platoonSpeed;
        data[9] = converter.bytes[0];
        data[10] = converter.bytes[1];
        data[11] = converter.bytes[2];
        data[12] = converter.bytes[3];
        converter.floatVal = innerPlatoonDistance;
        data[13] = converter.bytes[0];
        data[14] = converter.bytes[1];
        data[15] = converter.bytes[2];
        data[16] = converter.bytes[3];

        return data;
    }
};

template<>
struct Decoder<protocol::PlatoonMessage>
{
    protocol::PlatoonMessage operator()(const std::string & data)
    {
        using namespace protocol;

        if (data.size() < PlatoonMessage::SIZE)
            return PlatoonMessage{};

        VehicleId vehicleId{0};
        MessageType messageType{0};
        PlatoonId platoonId{0};
        PlatoonSpeed platoonSpeed{0};
        InnerPlatoonDistance innerPlatoonDistance{0};

        vehicleId += ((VehicleId) data[0]);
        vehicleId += ((VehicleId) data[1]) << 8;
        vehicleId += ((VehicleId) data[2]) << 16;
        vehicleId += ((VehicleId) data[3]) << 24;
        messageType += (MessageType) data[4];
        platoonId += ((PlatoonId) data[5]);
        platoonId += ((PlatoonId) data[6]) << 8;
        platoonId += ((PlatoonId) data[7]) << 16;
        platoonId += ((PlatoonId) data[8]) << 24;
        FloatConverter converter;
        converter.bytes[0] = data[9];
        converter.bytes[1] = data[10];
        converter.bytes[2] = data[11];
        converter.bytes[3] = data[12];
        platoonSpeed = converter.floatVal;
        converter.bytes[0] = data[13];
        converter.bytes[1] = data[14];
        converter.bytes[2] = data[15];
        converter.bytes[3] = data[16];
        innerPlatoonDistance = converter.floatVal;

        return PlatoonMessage{vehicleId, messageType, platoonId, platoonSpeed, innerPlatoonDistance};
    };
};

}
}

#endif //PROTOCOL_PLATOONCREATEMESSAGE_H
