//
// Created by philipp on 30.11.17.
//

#include <iostream>
#include "../include/LeaderVehicle.h"
#include "../include/Utils.h"

using namespace protocol;
using namespace std::chrono_literals;

constexpr networking::time::Duration LeaderVehicle::HEARTBEAT_TIMEOUT;
constexpr networking::time::Duration LeaderVehicle::BROADCAST_INTERVAL;

LeaderVehicle::Ptr LeaderVehicle::create(networking::Networking & net, VehicleId vehicleId)
{
    return std::make_shared<LeaderVehicle>(PrivateTag{}, net, vehicleId);
}

LeaderVehicle::LeaderVehicle(PrivateTag, networking::Networking & net, VehicleId vehicleId)
    : Vehicle(net, vehicleId, Role::LEADER)
{
    broadcastTimer = networking::time::Timer::create(net);
    heartbeatReceiver = networking::message::DatagramReceiver<PlatoonMessage>::create(net,
                                                                                      getOwnEndpoint().getUdpPort());
}

void LeaderVehicle::stop()
{
    Vehicle::stop();

    removeAllFollowers();

    stopBroadcasting();

    stopReceiveHeartbeats();
}

void LeaderVehicle::setPlatoonConfig(const PlatoonConfig & platoonConfig)
{
    this->platoonConfig = platoonConfig;
}

std::size_t LeaderVehicle::getNumFollowers()
{
    std::lock_guard<std::mutex> lock(followersMutex);
    return followers.size();
}

void LeaderVehicle::doCreatePlatoon()
{
    if (state != State::IDLE)
        return;

    setState(State::CREATING_PLATOON);

    // Generate a platoon id.
    platoonId = rand();

    Vehicle::doCreatePlatoon();
}

void LeaderVehicle::doRunPlatoon()
{
    if (state != State::CREATING_PLATOON)
        return;

    setState(State::RUNNING_PLATOON);

    startBroadcasting();

    startReceiveHeartbeats();

    onRunningPlatoonCallback();
}

void LeaderVehicle::doLeavePlatoon()
{
    if (state == State::IDLE)
        return;

    stop();

    setState(State::IDLE);

    onLeavingPlatoonCallback();
}

void LeaderVehicle::makePlatoonCreateRequest(const VehicleEndpoint & vehicleEndpoint)
{
    // We'll send a request indicating that we want to be the leader.
    auto self = shared_from_this();
    platoonCreateServiceClient->asyncCall(
        vehicleEndpoint.getHost(),
        vehicleEndpoint.getTcpPort(),
        PlatoonMessage::leaderRequest(getOwnEndpoint().getVehicleId(), platoonId),
        3s,
        [self, vehicleEndpoint](const auto & error, const auto & response)
        {
            if (self->state != State::CREATING_PLATOON &&
                self->state != State::RUNNING_PLATOON)
                return;

            if (error)
                return;

            if (response.getVehicleId() != vehicleEndpoint.getVehicleId() ||
                response.getMessageType() != messageTypes::ACCEPT_RESPONSE ||
                response.getPlatoonId() != self->platoonId)
                return;

            self->addFollower(vehicleEndpoint);

            self->doRunPlatoon();
        });
}

PlatoonMessage LeaderVehicle::makePlatoonCreateResponse(const networking::Resolver::Endpoint & clientEndpoint,
                                                        const PlatoonMessage & request)
{
    if (state != State::CREATING_PLATOON &&
        state != State::RUNNING_PLATOON)
        return PlatoonMessage::rejectResponse(getOwnEndpoint().getVehicleId());

    // Only consider this message if it is a follower request,
    // meaning that a follower wants us to be its leader.
    if (request.getMessageType() != messageTypes::FOLLOWER_REQUEST)
        return PlatoonMessage::rejectResponse(getOwnEndpoint().getVehicleId());

    auto followerId = request.getVehicleId();
    // TODO: change vehicle endpoint information!
    VehicleEndpoint followerEndpoint{
        followerId,
        clientEndpoint.ip,
        followerId + 10000,
        followerId + 10010};

    addFollower(followerEndpoint);

    doRunPlatoon();

    return PlatoonMessage::acceptResponse(getOwnEndpoint().getVehicleId(), platoonId);
}

bool LeaderVehicle::findFollower(VehicleId id)
{
    return followers.find(id) != followers.end();
}

void LeaderVehicle::addFollower(const VehicleEndpoint & endpoint)
{
    if (findFollower(endpoint.getVehicleId()))
        return;

    Follower follower;
    follower.endpoint = endpoint;
    follower.broadcastSender = networking::message::DatagramSender<PlatoonMessage>::create(net);
    follower.heartbeatTimer = networking::time::Timer::create(net);

    {
        std::lock_guard<std::mutex> lock(followersMutex);
        followers[endpoint.getVehicleId()] = follower;
    }

    startHeartbeatTimer(followers[endpoint.getVehicleId()]);
}

void LeaderVehicle::removeFollower(Follower & follower)
{
    stopHeartbeatTimer(follower);

    std::lock_guard<std::mutex> lock(followersMutex);
    followers.erase(follower.endpoint.getVehicleId());
}

void LeaderVehicle::startHeartbeatTimer(Follower & follower)
{
    auto followerId = follower.endpoint.getVehicleId();
    auto self = shared_from_this();
    follower.heartbeatTimer->startTimeout(
        HEARTBEAT_TIMEOUT,
        [self, followerId]
        {
            if (self->state != State::RUNNING_PLATOON)
                return;

            if (!self->findFollower(followerId))
                return;

            auto & follower = self->followers[followerId];
            self->removeFollower(follower);
        });
}

void LeaderVehicle::stopHeartbeatTimer(LeaderVehicle::Follower & follower)
{
    follower.heartbeatTimer->stop();
}

void LeaderVehicle::startReceiveHeartbeats()
{
    receiveHeartbeat();
}

void LeaderVehicle::receiveHeartbeat()
{
    auto self = shared_from_this();
    heartbeatReceiver->asyncReceive(
        networking::time::Duration::max(), // No timeout required here.
        [self](const auto & error, const auto & message, const auto & host, auto port)
        {
            if (self->state != State::RUNNING_PLATOON)
                return;

            auto followerId = message.getVehicleId();

            if (!error &&
                message.isValid() &&
                self->findFollower(followerId) &&
                message.getPlatoonId() == self->platoonId)
            {
                auto & follower = self->followers[followerId];

                switch (message.getMessageType())
                {
                    case messageTypes::LEAVE_PLATOON:
                        self->removeFollower(follower);
                        // If there are no more followers, we leave the platoon too.
                        if (self->followers.empty())
                            self->doLeavePlatoon();
                        break;

                    case messageTypes::HEARTBEAT:
                        // Reset the timer.
                        self->stopHeartbeatTimer(follower);
                        self->startHeartbeatTimer(follower);
                        break;
                }
            };

            // Receive next message.
            self->receiveHeartbeat();
        });
}

void LeaderVehicle::stopReceiveHeartbeats()
{
    heartbeatReceiver->stop();
}

void LeaderVehicle::startBroadcasting()
{
    auto self = shared_from_this();
    broadcastTimer->startPeriodicTimeout(
        BROADCAST_INTERVAL,
        [self]
        {
            if (self->state != State::RUNNING_PLATOON)
                return;

            PlatoonConfig configCopy = self->platoonConfig.load();

            for (auto & pair : self->followers)
            {
                auto & follower = pair.second;

                auto broadcastMsg = PlatoonMessage::broadcastMessage(
                    self->getOwnEndpoint().getVehicleId(),
                    self->platoonId,
                    configCopy.platoonSpeed,
                    configCopy.innerPlatoonDistance);

                // Send the broadcast message and forget about it.
                follower.broadcastSender->asyncSend(
                    broadcastMsg, follower.endpoint.getHost(), follower.endpoint.getUdpPort(), BROADCAST_INTERVAL,
                    [self](auto && ...)
                    {});
            }
        });
}

void LeaderVehicle::stopBroadcasting()
{
    broadcastTimer->stop();
}

void LeaderVehicle::removeAllFollowers()
{
    utils::foreachInMap(followers,
                        [this](auto & follower)
                        { this->removeFollower(follower); });
}

LeaderVehicle::Ptr LeaderVehicle::shared_from_this()
{
    return std::static_pointer_cast<LeaderVehicle>(Vehicle::shared_from_this());
}
